#!/usr/bin/python3
"""Load templates to django-secretary.

Usage: load_templates.py [options] <file>...
       load_templates.py -h | --help

Options:
  -h, --help                show this help message and exit
  --settings SETTINGS       The Python path to a settings module, e.g.
                            "myproject.settings.main". If this isn't provided, the
                            DJANGO_SETTINGS_MODULE environment variable will be
                            used.
"""
import os
from typing import Type

import django
from django.apps import apps
from django.conf import settings
from django.core.files.base import ContentFile
from django.db.transaction import atomic
from django.utils.timezone import utc
from docopt import docopt

try:
    import yaml
except ImportError:
    yaml = None


def save_asset(model: Type, fields: dict) -> None:
    """Save an asset."""
    # Collect asset data
    obj = None
    data = {}

    name = fields.get('name')
    if name is None:
        exit("Name is not defined for {}.".format(model.__name__))
    template = fields.get('template')
    if template is not None:
        try:
            data['template'] = model.template.get_queryset().get(**template)
        except model.DoesNotExist:
            exit("Template for {} not found.".format(name))

    for key in ('mime_type', ):
        if key in fields:
            data[key] = fields[key]

    if model.objects.filter(name=name).exists():
        print("Asset {} already exists, skipping.".format(name))
        return
    data['name'] = name

    # Save asset
    obj = model(**data)
    with open(os.path.join('assets', fields.get('content', name)), 'rb') as content:
        obj.content.save(os.path.basename(name), ContentFile(content.read()))
    obj.save()
    print("Asset {} created.".format(name))


def _normalize_content(content: str) -> str:
    return content.strip().replace('\r', '')


def save_template(model: Type, fields: dict) -> None:
    """Save a template."""
    # Collect template data
    obj = None
    data = {'is_active': True}
    uuid = fields.get('uuid')
    if uuid is not None:
        obj = model.objects.filter(uuid=uuid).first()
        if obj is not None:
            print("Template {} already exists, skipping.".format(uuid))
            return
        data['uuid'] = uuid

    name = fields.get('name')
    if name is None:
        exit("Name is not defined for {}.".format(model.__name__))
    data['name'] = name

    parent = fields.get('parent')
    if parent is not None:
        try:
            data['parent'] = model.objects.get(**parent)
        except model.MultipleObjectsReturned:
            # Search active templates only
            try:
                data['parent'] = model.objects.get(**parent, is_active=True)
            except model.DoesNotExist:
                exit("Parent of {} not found.".format(name))

    for key in ('create_datetime', 'is_active', 'signer'):
        if key in fields:
            data[key] = fields[key]

    with open(os.path.join('templates', fields.get('content', name))) as file:
        content = file.read()

    # Try to find object in database.
    assert obj is None
    obj = model.objects.filter(**data).first()
    if obj is not None:
        if _normalize_content(obj.content.read().decode()) == _normalize_content(content):
            # It's the same model, skip.
            obj.content.close()
            print("Template {} is at the latest version, skipping.".format(obj))
            return
        else:
            # It's a different model. Deactivate it and create a new one below.
            obj.is_active = False
            obj.save()
            print("Template {} deactivated.".format(obj))
    elif data['is_active']:
        # Object not found. Try to find active template by name only to prevent duplicate problems.
        obj = model.objects.filter(name=data['name'], is_active=True).first()
        if obj is not None:
            # It's a different model. Deactivate it and create a new one below.
            obj.is_active = False
            obj.save()
            print("Template {} deactivated.".format(obj))

    # Save new template
    obj = model(**data)
    obj.content.save(name, ContentFile(content))
    obj.save()
    # Field `create_datetime` is always updated. Force the value if needed.
    if 'create_datetime' in data:
        create_datetime = data['create_datetime']
        if settings.USE_TZ:
            create_datetime = create_datetime.astimezone(utc)
        obj.create_datetime = create_datetime
        obj.save()
    print("Template {} created.".format(obj))


def save_template_collection(model: Type, fields: dict) -> None:
    """Save a template collection."""
    # Collect collection data
    obj = None
    data = {}

    name = fields.get('name')
    if name is None:
        exit("Name is not defined for {}.".format(model.__name__))
    if model.objects.filter(name=name).exists():
        print("Collection {} already exists, skipping.".format(name))
        return
    data['name'] = name
    data['label'] = fields.get('label', name)

    # Save collection
    obj = model(**data)
    obj.save()
    tag_model = obj.tags.model
    for tag_name in fields.get('tags', []):
        tag, _ = tag_model.objects.get_or_create(name=tag_name)
        obj.tags.add(tag)
    print("Collection {} created.".format(name))


def save_obj(data: dict):
    """Save single object."""
    model_name = data.get('model')
    if model_name is None:
        exit("Model is not defined in {}.".format(data))
    model = apps.get_model(model_name)
    fields = data.get('fields')
    if fields is None:
        exit("Fields are not defined for {}.".format(model_name))

    if model_name.lower() == 'django_secretary.asset':
        save_asset(model, fields)
    elif model_name.lower() == 'django_secretary.template':
        save_template(model, fields)
    elif model_name.lower() == 'django_secretary.templatecollection':
        save_template_collection(model, fields)
    else:
        exit("Model {} is not supported.".format(model_name))


def load_file(filename: str):
    """Load and save objects from a file."""
    print("Loading {}".format(filename))
    with atomic():
        with open(filename) as fixture:
            for document in yaml.safe_load_all(fixture):
                for data in document:
                    save_obj(data)


def main():
    options = docopt(__doc__)
    if options['--settings']:
        os.environ['DJANGO_SETTINGS_MODULE'] = options['--settings']
    django.setup()

    if not yaml:
        exit("yaml library not found")

    for filename in options['<file>']:
        load_file(filename)


if __name__ == '__main__':
    main()
